/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.crop.model;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import javax.swing.event.UndoableEditEvent;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import org.openide.awt.UndoRedo;
import org.openide.util.Exceptions;

/**
 * PropertyChangeSupport for the crop model
 *
 * @author Merijn Zengers
 */
public class CropBean implements Serializable {

    public static final String PROP_NAME = "name";
    private String name;
    private transient PropertyChangeSupport propertyChangeSupport;
    private transient List<UndoRedo.Manager> undoRedoManagers;
    public transient BeanState currentState = BeanState.NORMAL;

    public enum BeanState {

        NORMAL, UNDO
    };

    public String getName() {
        return name;
    }

    public void setName(String name) {
        String old = this.name;
        this.name = name;
        getPropertyChangeSupport().firePropertyChange(PROP_NAME, old, name);
        fireUndoableEvent(PROP_NAME, this, old, name);
    }

    public void addPropertyChangeListener(PropertyChangeListener pcl) {
        getPropertyChangeSupport().addPropertyChangeListener(pcl);
    }

    public void addPropertyChangeListener(String propertyName, PropertyChangeListener pcl) {
        getPropertyChangeSupport().addPropertyChangeListener(propertyName, pcl);
    }

    public void removePropertyChangeListener(PropertyChangeListener pcl) {
        getPropertyChangeSupport().removePropertyChangeListener(pcl);
    }

    public void removePropertyChangeListener(String propertyName, PropertyChangeListener pcl) {
        getPropertyChangeSupport().removePropertyChangeListener(propertyName, pcl);
    }

    PropertyChangeSupport getPropertyChangeSupport() {
        if (propertyChangeSupport == null) {
            propertyChangeSupport = new PropertyChangeSupport(this);
        }
        return propertyChangeSupport;
    }

    private BeanState getCurrentState() {
        return currentState;
    }

    private void setCurrentState(BeanState currentState) {
        this.currentState = currentState;
    }

    private List<UndoRedo.Manager> getUndoRedoManagers() {
        if (undoRedoManagers == null) {
            undoRedoManagers = new ArrayList<UndoRedo.Manager>();
        }
        return undoRedoManagers;
    }

    public void addUndoableEditListener(UndoRedo.Manager manager) {
        getUndoRedoManagers().add(manager);
    }

    public void removeUndoableEditListener(UndoRedo.Manager manager) {
        getUndoRedoManagers().remove(manager);
    }

    public void fireUndoableEvent(String property, CropBean source, Object oldValue, Object newValue) {
        if (getCurrentState() != BeanState.UNDO) {
            for (UndoRedo.Manager manager : getUndoRedoManagers()) {
                if (manager != null) {
                    ReUndoableEdit reUndoableEdit = new ReUndoableEdit(property, source, oldValue, newValue);
                    UndoableEditEvent undoableEditEvent = new UndoableEditEvent(this, reUndoableEdit);


                    assert undoableEditEvent != null;

                    manager.undoableEditHappened(undoableEditEvent);
                }
            }
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final CropBean other = (CropBean) obj;
        if ((this.name == null) ? (other.name != null) : !this.name.equals(other.name)) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = 3;
        hash = 23 * hash + (this.name != null ? this.name.hashCode() : 0);
        return hash;
    }

    /*
     * Custom UndoableEdit extending abstractUndoableEdit Custom implemented
     * Undo & Redo functionality Will call setters on bean using reflection when
     * undo or redo is called.
     *
     *
     * @Author Gerben Feenstra & Merijn Zengers
     *
     */
    private class ReUndoableEdit extends AbstractUndoableEdit {

        private Object oldValue;
        private Object newValue;
        private CropBean source;
        private String property;

        public ReUndoableEdit(String property, CropBean source, Object oldValue, Object newValue) {
            super();
            this.oldValue = oldValue;
            this.newValue = newValue;
            this.source = source;
            this.property = property;
        }

        private Method getMethod() {
            try {
                return source.getClass().getMethod("set" + property.replaceFirst(property.substring(0, 1), property.substring(0, 1).toUpperCase()), newValue.getClass());
            } catch (    NoSuchMethodException | SecurityException ex) {
                Exceptions.printStackTrace(ex);
            }
            return null;
        }

        @Override
        public void undo() throws CannotUndoException {
            try {
                super.undo();
                Method method = getMethod();
                setCurrentState(BeanState.UNDO);
                method.invoke(source, oldValue);
                setCurrentState(BeanState.NORMAL);
            } catch (    SecurityException | IllegalAccessException | IllegalArgumentException | InvocationTargetException ex) {
                Exceptions.printStackTrace(ex);
            }
        }

        @Override
        public void redo() throws CannotRedoException {
            try {
                super.redo();
                Method method = getMethod();
                setCurrentState(BeanState.UNDO);
                method.invoke(source, newValue);
                setCurrentState(BeanState.NORMAL);
            } catch (    IllegalAccessException | IllegalArgumentException | InvocationTargetException ex) {
                Exceptions.printStackTrace(ex);
            }
        }
    }
}
